#!C:\b\abs_c1vy7lsvi7\croot\gdal-split_1680120439677\_h_env\python.exe
# EASY-INSTALL-SCRIPT: 'GDAL==3.6.2','gdalmove.py'
__requires__ = 'GDAL==3.6.2'
__import__('pkg_resources').run_script('GDAL==3.6.2', 'gdalmove.py')
